module ngmap where

open import Data.Unit hiding (_≟_)
open import Data.Product
open import Data.Sum
open import Data.List hiding ([_]; replicate)
open import Data.Vec  hiding ([_]; _⊛_)
open import Data.Vec1
open import Data.Bool hiding (_≟_)
open import Data.Unit1
open import Data.Function hiding (_⟨_⟩_)
open import Data.Nat

open import univ
open import local-lib
open import aritygen
open import nvec-map

NGmap : {n : ℕ} → Vec₁ Set (suc n) → Set1
NGmap {n} v = ⇑ (arrTy {n} v)

postulate error : (A : Set) → A

defNat : (n : ℕ) → NGmap ⟨ ⋆ ⟩ (repeat₁ {suc n} ℕ)
defNat zero           = ↑ zero            -- arbitrary ℕ 
defNat (suc zero)     = ↑ \ x → x         -- return what was given
defNat (suc (suc n))  = 
  ↑ \ x → \ y → if eqNat x y  then (↓ (defNat (suc n))) y
                              else error _

defUnit : (n : ℕ) → NGmap ⟨ ⋆ ⟩ (repeat₁ {suc n} ⊤)
defUnit zero     = ↑ tt
defUnit (suc n)  = ↑ \ x → ↓ (defUnit n)

defPairAux  : (n : ℕ) 
         → (As : Vec₁ Set (suc n)) → NGmap As
         → (Bs : Vec₁ Set (suc n)) → NGmap Bs 
         → NGmap (repeat₁ {suc n} _×_ ⊛₁ As ⊛₁ Bs)
defPairAux zero     (A ∷ [])   (↑ a)  (B ∷ [])   (↑ b)  = ↑ (a , b)
defPairAux (suc n)  (A1 ∷ As)  (↑ a)  (B1 ∷ Bs)  (↑ b)  =
 ↑ \ p →
    ↓ (defPairAux n As (↑ (a (proj₁ p))) Bs (↑ (b (proj₂ p))))

defPair  : (n : ℕ) 
         → NGmap ⟨ ⋆ ⇒ ⋆ ⇒ ⋆ ⟩ (repeat₁ {suc n} _×_) 
defPair n = k-curry (⋆ ⇒ ⋆ ⇒ ⋆) (defPairAux n)

defSumLeft  : (n : ℕ) 
            → (As : Vec₁ Set (suc n)) → arrTy As
            → (Bs : Vec₁ Set (suc n))
            → arrTy (repeat₁ _⊎_ ⊛₁ As ⊛₁ Bs)
defSumLeft zero     (A ∷ [])   a  (B ∷ [])   = inj₁ a
defSumLeft (suc n)  (A1 ∷ As)  a  (B1 ∷ Bs)  = f 
  where 
    f : A1 ⊎ B1 → arrTy (repeat₁ _⊎_ ⊛₁ As ⊛₁ Bs)
    f (inj₁ a1)  = defSumLeft n As (a a1) Bs 
    f (inj₂ b1)  = defSumLeft n As (a (error A1)) Bs

defSumRight  : (n : ℕ)
             → (As : Vec₁ Set (suc n))
             → (Bs : Vec₁ Set (suc n)) → arrTy Bs
             → arrTy (repeat₁ _⊎_ ⊛₁ As ⊛₁ Bs)
defSumRight zero     (A ∷ [])   (B ∷ [])   b  = inj₂ b
defSumRight (suc n)  (A1 ∷ As)  (B1 ∷ Bs)  b  = f 
  where 
    f : A1 ⊎ B1 → arrTy (repeat₁ _⊎_ ⊛₁ As ⊛₁ Bs)
    f (inj₁ a1)  = defSumRight n As Bs (b (error B1))
    f (inj₂ b1)  = defSumRight n As Bs (b b1)

defSumAux : (n : ℕ) 
         → (As : Vec₁ Set (suc n)) → NGmap As
         → (Bs : Vec₁ Set (suc n)) → NGmap Bs
         → NGmap (repeat₁ {suc n} _⊎_ ⊛₁ As ⊛₁ Bs)
defSumAux zero (A ∷ []) (↑ a) (B ∷ []) (↑ b) = 
  ↑ (inj₂ b)    --arbitrary
defSumAux (suc n) (A1 ∷ As) (↑ a) (B1 ∷ Bs) (↑ b) = ↑ f
  where 
    f : A1 ⊎ B1 → arrTy (repeat₁ _⊎_ ⊛₁ As ⊛₁ Bs)
    f (inj₁ a1)  = defSumLeft  n As (a a1) Bs
    f (inj₂ b1)  = defSumRight n As Bs (b b1)

defSum  : (n : ℕ) 
        → NGmap ⟨ ⋆ ⇒ ⋆ ⇒ ⋆ ⟩ (repeat₁ {suc n}  _⊎_) 
defSum n = k-curry (⋆ ⇒ ⋆ ⇒ ⋆) (defSumAux n)

ngmap-const : {n : ℕ} → TyConstEnv (NGmap {n})
ngmap-const {n} Nat   = defNat n
ngmap-const {n} Unit  = defUnit n
ngmap-const {n} Prod  = defPair n
ngmap-const {n} Sum   = defSum n

ngmap  : (n : ℕ) → {k : Kind} → (e : Ty k) 
       → NGmap ⟨ k ⟩ (repeat₁ {suc n} ⌊ e ⌋)
ngmap n e = ngen e ngmap-const

option-map0 : { A : Set } → A → Option A
option-map0 f = ↓ (ngmap 0 option (↑ f))

option-map1  : { A B : Set } → (A → B) 
            → Option A → Option B
option-map1 f = ↓ (ngmap 1 option (↑ f))

option-map2  : { A B C : Set } → (A → B → C) 
            → Option A → Option B → Option C
option-map2 f = ↓ (ngmap 2 option (↑ f))

pair : Ty (⋆ ⇒ ⋆ ⇒ ⋆)
pair = Con Prod

pair-map0 : { A B : Set } → A → B → A × B
pair-map0 x y = ↓ (ngmap 0 (Con Prod) (↑ x) (↑ y))

pair-map1  : { A1 B1 A2 B2 : Set } 
           → (A1 → B1) → (A2 → B2) 
           → A1 × A2 → B1 × B2
pair-map1 x y = ↓ (ngmap 1 (Con Prod) (↑ x) (↑ y))

pair-map2  : { A1 B1 C1 A2 B2 C2 : Set }
           → (A1 → B1 → C1) → (A2 → B2 → C2)
           → A1 × A2 → B1 × B2 → C1 × C2
pair-map2 f1 f2 = ↓ (ngmap 2 (Con Prod) (↑ f1) (↑ f2))

-- Tests
t1 = option-map0 zero 

t2 = option-map1 suc (inj₂ zero) 

t3 = pair-map1 suc (\ x → x) (zero , zero)

