{-# OPTIONS --type-in-type #-}

module aritygen where

-- Arity generic definition of generic functions

open import Data.Unit
open import Data.Product
open import Data.Nat
open import Data.Sum
open import Data.List
open import Data.Vec1
open import Data.Bool
open import Relation.Binary.PropositionalEquality1

open import univ
open import local-lib
open import nvec-map

_⟨_⟩_ : {n : ℕ} → (Vec₁ Set (suc n) → Set1) → (k : Kind) → Vec₁ ⟦ k ⟧ (suc n) → Set1
b ⟨ ⋆ ⟩ v   = b v
b ⟨ k1 ⇒ k2 ⟩ v = ∀⇒ \ (a : Vec₁ ⟦ k1 ⟧ _) → (b ⟨ k1 ⟩ a → b ⟨ k2 ⟩ (v ⊛₁ a))

_⟨⟨_⟩⟩_ : {n : ℕ} → (Vec₁ Set (suc n) → Set1) → (k : Kind) → Vec₁ ⟦ k ⟧ (suc n) → Set1
b ⟨⟨ ⋆ ⟩⟩ v   = b v
b ⟨⟨ k1 ⇒ k2 ⟩⟩ v =  (a : Vec₁ ⟦ k1 ⟧ _) → (b ⟨ k1 ⟩ a → b ⟨⟨ k2 ⟩⟩ (v ⊛₁ a))

k-curry : (k : Kind)  → {n : ℕ} → {b : Vec₁ Set (suc n) → Set1} 
          → { Vs : Vec₁ ⟦ k ⟧ (suc n) } 
          → b ⟨⟨ k ⟩⟩ Vs  → b ⟨ k ⟩ Vs
k-curry ⋆ x = x
k-curry (k1 ⇒ k2){n}{b}{Vs} x = 
  \⇒ {suc n}{ ⟦ k1 ⟧ } 
     {\ (X : Vec₁ ⟦ k1 ⟧ (suc n)) → b ⟨ k1 ⟩ X → b ⟨ k2 ⟩ (Vs ⊛₁ X)} 
      (\ As y → k-curry k2 {n}{b}{Vs ⊛₁ As} (x As y))
  
     

-- we can make this a record instead of a function, but I like 
-- it better this way. This file is independent of the actual 
-- constants of the language.
TyConstEnv :  {n : ℕ} → (b : Vec₁ Set (suc n) → Set1) → Set1 
TyConstEnv b = 
  {k : Kind} (c : Const k) → b ⟨ k ⟩ repeat₁ ⌊ Con c ⌋

data NGEnv  {n : ℕ} (b : Vec₁ Set (suc n) → Set1) 
            : Ctx → Set1 where
   NNil   : NGEnv b []
   NCons  : {k : Kind}  {G : Ctx}
          → (a : Vec₁ ⟦ k ⟧ (suc n))
          → b ⟨ k ⟩ a 
          → NGEnv b G 
          → NGEnv b (k ∷ G)

interp∗  : ∀ {G k n} → Typ G k → Vec₁ (Env G) n 
    → Vec₁ ⟦ k ⟧ n
interp∗ t vs = repeat₁ (interp t) ⊛₁ vs

transpose  :  {n : ℕ} {b : Vec₁ Set (suc n) → Set1} 
              {G : Ctx}
           → NGEnv b G → Vec₁ (Env G) (suc n)
transpose NNil            = repeat₁ []
transpose (NCons a _ nge)  = 
  (repeat₁ _∷_) ⊛₁ a ⊛₁ (transpose nge)

≡-KIT  :  {n : ℕ} {b : Vec₁ Set (suc n) → Set1} 
          { k : Kind } {t1 t2 : Vec₁ ⟦ k ⟧ (suc n)} 
        → t1 ≡₁ t2 
        → b ⟨ k ⟩ t1 
        → b ⟨ k ⟩ t2
≡-KIT refl x = x

c1  : {n : ℕ} { k : Kind } {G : Ctx}
    → (a : Vec₁ ⟦ k ⟧ n)
    → (envs : Vec₁ (Env G) n) 
    → a ≡₁ interp∗ ( Var VZ ) (repeat₁ _∷_ ⊛₁ a ⊛₁ envs)
c1 {zero}  []        []          = refl
c1 {suc n} (t ∷ ts)  ( x ∷ xs )  = ≡-tail₁ (c1 {n} ts xs)

c2  : { n : ℕ } { k k' : Kind } {G : Ctx}
    → (x : TyVar G k')
    → (t1 : Vec₁ ⟦ k ⟧ n) 
    → (envs : Vec₁ (Env G) n) 
    →  interp∗ ( Var x ) envs ≡₁ 
       interp∗ ( Var (VS x) ) (repeat₁ _∷_ ⊛₁ t1 ⊛₁ envs)
c2 {zero}  x []       []        = refl
c2 {suc n} x (t ∷ ts) (y ∷ ys)  = ≡-tail₁ (c2 x ts ys)

nLookup  :  {n : ℕ} {b : Vec₁ Set (suc n) → Set1}
            {k : Kind} {G : Ctx} 
         → (v : TyVar G k) 
         → (nge : NGEnv b G) 
         → b ⟨ k ⟩ (interp∗ ( Var v ) (transpose nge))
nLookup {n} {b} {k} VZ  (NCons a e nge) = 
    ≡-KIT (c1 a (transpose nge))     e
nLookup (VS x) (NCons a _ nge) = 
    ≡-KIT (c2 x a (transpose nge))   (nLookup x nge)

c3 : {n : ℕ}{k k' : Kind}{G : Ctx}
    → (t : Typ (k' ∷ G) k)
    → (envs : Vec₁ (Env G) n) 
    → (as : Vec₁ ⟦ k' ⟧ n) 
    → (interp∗ ( t ) (repeat₁ _∷_ ⊛₁ as ⊛₁ envs))
    ≡₁ (interp∗ ( Lam t ) envs) ⊛₁ as
c3 {zero} t [] [] = refl 
c3 {suc n} t (a ∷ as) (b ∷ bs) = ≡-tail₁ (c3 t as bs)

c4 : {n : ℕ}{k1 k2 : Kind} {G : Ctx}
     → (t1 : Typ G (k1 ⇒ k2))
     → (t2 : Typ G k1)
     → (envs : Vec₁ (Env G) n) 
     →    (interp∗ ( t1 ) envs) ⊛₁ (interp∗ ( t2 ) envs)
      ≡₁  interp∗ ( App t1 t2 ) envs
c4 {zero}  _  _  []       = refl
c4 {suc n} t1 t2 (a ∷ as) = ≡-tail₁ (c4 t1 t2 as)

c5 : {n : ℕ}{k : Kind}{G : Ctx}
  → (c : Const k)
  → (envs : Vec₁ (Env G) n) 
  → repeat₁ ⌊ Con c ⌋ ≡₁ interp∗ ( Con c ) envs
c5 {zero}  _ []       = refl
c5 {suc n} c (a ∷ as) = ≡-tail₁ (c5 c as)

ngen-open : {n : ℕ}{b : Vec₁ Set (suc n) → Set1}{G : Ctx} {k : Kind} → 
             (t : Typ G k) → (ve : NGEnv b G) → 
             (ce : TyConstEnv b) →
             b ⟨ k ⟩ ( interp∗ t (transpose ve))
ngen-open (Var x) ve ce = nLookup x ve
ngen-open{n}{b} (Lam {k1 = k1} t) ve ce = 
  \⇒ (\ (a : Vec₁ ⟦ k1 ⟧ (suc n)) (nwt : b ⟨ k1 ⟩ a) → 
  ≡-KIT (c3 t (transpose ve) a) 
  (ngen-open t (NCons a nwt ve) ce))
ngen-open{n}{b}{G} (App {k1 = k1}{k2 = k2} t1 t2) ve ce = 
  ≡-KIT (c4 t1 t2 (transpose ve)) 
  (/⇒ (suc n) {⟦ k1 ⟧}{λ a → ( b ⟨ k1 ⟩ a → 
        ( b ⟨ k2 ⟩ ( (interp∗ t1 (transpose ve)) ⊛₁ a)))}
      (ngen-open {n}{b}{G}{k1 ⇒ k2} t1 ve ce)
      (interp∗  t2 (transpose ve)) (ngen-open t2 ve ce))
ngen-open (Con c) ve ce = ≡-KIT (c5 c (transpose ve)) (ce c) 

c6 : { n : ℕ } { A B : Set1 } { f : A → B } { x : A } →
  (repeat₁{n} f) ⊛₁ repeat₁ x ≡₁ repeat₁ (f x)
c6 {zero} = refl
c6 {suc n} = ≡-tail₁ c6

ngen : {n : ℕ}{b : Vec₁ Set (suc n) → Set1}{k : Kind} → 
       (t : Ty k) → (TyConstEnv b) → b ⟨ k ⟩ (repeat₁ ⌊ t ⌋)
ngen{n}{b}{k} t ce = ≡-KIT {n}{b}{k} c6 (ngen-open t NNil ce)


-- the arity2 map example from section 3.2

GMap : Vec₁ Set 2 → Set
GMap (A ∷ B ∷ []) = A → B

gmap-const : TyConstEnv GMap
gmap-const Nat   = \ x → x
gmap-const Unit  = \ x → x 
gmap-const Prod  = \ fa fb x → 
  ( fa (proj₁ x) , fb (proj₂ x) ) 
gmap-const Sum   = \ fa fb x → g fa fb x
      where g  : {A1 B1 A2 B2 : Set} 
               → (A1 → B1) → (A2 → B2) 
               → A1 ⊎ A2 → B1 ⊎ B2
            g  fa fb (inj₁ xa) = inj₁ (fa xa)
            g  fa fb (inj₂ xb) = inj₂ (fb xb)

gmap  : {k : Kind} → (t : Ty k) 
      → GMap ⟨ k ⟩ ( ⌊ t ⌋ ∷ ⌊ t ⌋ ∷ [])
gmap t = ngen t gmap-const

option-map  : { A B : Set } → (A → B) 
            → Option A → Option B
option-map = gmap option

