module univ where

-- Defines the basic universe for generic programming in Agda as 
-- well as the interpretation of kinds and types.
-- Does not support datatype views.

open import Data.Unit
open import Data.Product
open import Data.Nat
open import Data.Sum
open import Data.List
open import Data.Vec
open import Data.Vec1
open import Data.Bool

open import local-lib
infixr 50 _⇒_

data Kind : Set where 
  ⋆    : Kind 
  _⇒_  : Kind → Kind → Kind

⟦_⟧        : Kind → Set1
⟦ ⋆ ⟧      = Set
⟦ a ⇒ b ⟧  = ⟦ a ⟧ → ⟦ b ⟧

data Const : Kind → Set where
  Nat   : Const ⋆
  Unit  : Const ⋆
  Sum   : Const (⋆ ⇒ ⋆ ⇒ ⋆) 
  Prod  : Const (⋆ ⇒ ⋆ ⇒ ⋆)

interp-c : ∀ {k} → Const k → ⟦ k ⟧
interp-c Unit     = ⊤
interp-c Nat      = ℕ
interp-c Sum      = _⊎_
interp-c Prod     = _×_

Ctx : Set
Ctx = List Kind

data TyVar : Ctx → Kind → Set where
  VZ   : ∀ {G k} → TyVar (k ∷ G) k
  VS   : ∀ {G k' k} → TyVar G k → TyVar (k' ∷ G) k 

data Typ : Ctx → Kind → Set where 
  Var  : ∀ {G k} → TyVar G k → Typ G k 
  Lam  : ∀ {G k1 k2} → Typ (k1 ∷ G) k2 
       → Typ G (k1 ⇒ k2)
  App  : ∀ {G k1 k2} → Typ G (k1 ⇒ k2) → Typ G k1
       → Typ G k2
  Con : ∀ {G k} → Const k → Typ G k

Ty : Kind → Set
Ty = Typ []

data Env : List Kind → Set1 where
  []   : Env [] 
  _∷_  : ∀ {k G} → ⟦ k ⟧ → Env G → Env (k ∷ G)

sLookup : ∀ {k G} → TyVar G k → Env G → ⟦ k ⟧
sLookup VZ      (v ∷ G) = v
sLookup (VS x)  (v ∷ G) = sLookup x G

interp : ∀ {k G} → Typ G k → Env G → ⟦ k ⟧
interp (Var x) e        = sLookup x e
interp (Lam t) e        = \ y → interp t (y ∷ e)
interp (App t1 t2) e    = (interp t1 e) (interp t2 e)
interp (Con c) e        = interp-c c

⌊_⌋ : ∀ {k} → Ty k → ⟦ k ⟧
⌊ t ⌋ = interp t []

Option : Set → Set
Option = \ A → ⊤ ⊎ A

option : Ty (⋆ ⇒ ⋆) 
option = 
  Lam (App (App (Con Sum) (Con Unit)) (Var VZ))

